@@ -19,7 +19,6 @@ from __future__ import print_function
 import argparse
 import collections
 import os
-import sys
 
 from oslo_utils import strutils
 import six
@@ -31,13 +30,21 @@ from cinderclient import exceptions
 from cinderclient import shell_utils
 from cinderclient import utils
 
-from cinderclient.v2.shell import *  # flake8: noqa
+from cinderclient.v2.shell import *  # noqa
 
 FILTER_DEPRECATED = ("This option is deprecated and will be removed in "
                      "newer release. Please use '--filters' option which "
                      "is introduced since 3.33 instead.")
 
 
+class AppendFilters(argparse.Action):
+
+    filters = []
+
+    def __call__(self, parser, namespace, values, option_string):
+        AppendFilters.filters.append(values[0])
+
+
 @api_versions.wraps('3.33')
 @utils.arg('--resource',
            metavar='<resource>',
@@ -53,6 +60,7 @@ def do_list_filters(cs, args):
 
 
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.52',
@@ -67,8 +75,8 @@ def do_type_list(cs, args):
     # pylint: disable=function-redefined
     search_opts = {}
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
     vtypes = cs.volume_types.list(search_opts=search_opts)
     shell_utils.print_volume_type_list(vtypes)
 
@@ -84,6 +92,7 @@ def do_type_list(cs, args):
             mode="w"):
         for vtype in vtypes:
             cs.volume_types.write_to_completion_cache('name', vtype.name)
+    AppendFilters.filters = []
 
 
 @utils.arg('--all-tenants',
@@ -133,6 +142,7 @@ def do_type_list(cs, args):
                   'Valid keys: %s. '
                   'Default=None.') % ', '.join(base.SORT_KEY_VALUES)))
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -164,8 +174,8 @@ def do_backup_list(cs, args):
     }
 
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     total_count = 0
     if show_count:
@@ -206,12 +216,14 @@ def do_backup_list(cs, args):
         for backup in backups:
             if backup.name is not None:
                 cs.backups.write_to_completion_cache('name', backup.name)
+    AppendFilters.filters = []
 
 
 @utils.arg('--detail',
            action='store_true',
            help='Show detailed information about pools.')
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -224,8 +236,8 @@ def do_get_pools(cs, args):
     # pylint: disable=function-redefined
     search_opts = {}
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
     if cs.api_version >= api_versions.APIVersion("3.33"):
         pools = cs.volumes.get_pools(args.detail, search_opts)
     else:
@@ -239,6 +251,7 @@ def do_get_pools(cs, args):
         if args.detail:
             backend.update(info['capabilities'])
         utils.print_dict(backend)
+    AppendFilters.filters = []
 
 
 RESET_STATE_RESOURCES = {'volume': utils.find_volume,
@@ -324,14 +337,6 @@ RESET_STATE_RESOURCES = {'volume': utils.find_volume,
                 'Use the show command to see which fields are available. '
                 'Unavailable/non-existent fields will be ignored. '
                 'Default=None.')
-@utils.arg('--sort_key',
-           metavar='<sort_key>',
-           default=None,
-           help=argparse.SUPPRESS)
-@utils.arg('--sort_dir',
-           metavar='<sort_dir>',
-           default=None,
-           help=argparse.SUPPRESS)
 @utils.arg('--sort',
            metavar='<key>[:<direction>]',
            default=None,
@@ -346,6 +351,7 @@ RESET_STATE_RESOURCES = {'volume': utils.find_volume,
            metavar='<tenant>',
            help='Display information from single tenant (Admin only).')
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -388,8 +394,8 @@ def do_list(cs, args):
         'group_id': getattr(args, 'group_id', None),
     }
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     # If unavailable/non-existent fields are specified, these fields will
     # be removed from key_list at the print_list() during key validation.
@@ -398,24 +404,15 @@ def do_list(cs, args):
         for field_title in args.fields.split(','):
             field_titles.append(field_title)
 
-    # --sort_key and --sort_dir deprecated in kilo and is not supported
-    # with --sort
-    if args.sort and (args.sort_key or args.sort_dir):
-        raise exceptions.CommandError(
-            'The --sort_key and --sort_dir arguments are deprecated and are '
-            'not supported with --sort.')
-
     total_count = 0
     if show_count:
         search_opts['with_count'] = args.with_count
         volumes, total_count = cs.volumes.list(
             search_opts=search_opts, marker=args.marker,
-            limit=args.limit, sort_key=args.sort_key,
-            sort_dir=args.sort_dir, sort=args.sort)
+            limit=args.limit, sort=args.sort)
     else:
         volumes = cs.volumes.list(search_opts=search_opts, marker=args.marker,
-                                  limit=args.limit, sort_key=args.sort_key,
-                                  sort_dir=args.sort_dir, sort=args.sort)
+                                  limit=args.limit, sort=args.sort)
     shell_utils.translate_volume_keys(volumes)
 
     # Create a list of servers to which the volume is attached
@@ -451,7 +448,7 @@ def do_list(cs, args):
         if search_opts['all_tenants']:
             key_list.insert(1, 'Tenant ID')
 
-    if args.sort_key or args.sort_dir or args.sort:
+    if args.sort:
         sortby_index = None
     else:
         sortby_index = 0
@@ -459,6 +456,7 @@ def do_list(cs, args):
                      sortby_index=sortby_index)
     if show_count:
         print("Volume in total: %s" % total_count)
+    AppendFilters.filters = []
 
 
 @utils.arg('entity', metavar='<entity>', nargs='+',
@@ -607,12 +605,6 @@ def do_reset_state(cs, args):
            help='Scheduler hint, similar to nova. Repeat option to set '
                 'multiple hints. Values with the same key will be stored '
                 'as a list.')
-@utils.arg('--allow-multiattach',
-           dest='multiattach',
-           action="store_true",
-           help=('Allow volume to be attached more than once. (DEPRECATED)'
-                 ' Default=False'),
-           default=False)
 @utils.arg('--poll',
            action="store_true",
            help=('Wait for volume creation until it completes.'))
@@ -667,7 +659,6 @@ def do_create(cs, args):
                                imageRef=image_ref,
                                metadata=volume_metadata,
                                scheduler_hints=hints,
-                               multiattach=args.multiattach,
                                backup_id=backup_id)
 
     info = dict()
@@ -681,8 +672,11 @@ def do_create(cs, args):
 
     if args.poll:
         timeout_period = os.environ.get("POLL_TIMEOUT_PERIOD", 3600)
-        shell_utils._poll_for_status(cs.volumes.get, volume.id, info, 'creating', ['available'],
-                timeout_period, cs.client.global_request_id, cs.messages)
+        shell_utils._poll_for_status(
+            cs.volumes.get, volume.id, info, 'creating', ['available'],
+            timeout_period, cs.client.global_request_id, cs.messages)
+        volume = cs.volumes.get(volume.id)
+        info.update(volume._info)
 
     utils.print_dict(info)
 
@@ -753,10 +747,23 @@ def do_summary(cs, args):
 
 
 @api_versions.wraps('3.11')
+@utils.arg('--filters',
+           action=AppendFilters,
+           type=six.text_type,
+           nargs='*',
+           start_version='3.52',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Admin only.")
 def do_group_type_list(cs, args):
     """Lists available 'group types'. (Admin only will see private types)"""
-    gtypes = cs.group_types.list()
+    search_opts = {}
+    # Update search option with `filters`
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
+    gtypes = cs.group_types.list(search_opts=search_opts)
     shell_utils.print_group_type_list(gtypes)
+    AppendFilters.filters = []
 
 
 @api_versions.wraps('3.11')
@@ -1331,6 +1338,7 @@ def do_manageable_list(cs, args):
            default=utils.env('ALL_TENANTS', default=None),
            help='Shows details for all tenants. Admin only.')
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -1344,14 +1352,29 @@ def do_group_list(cs, args):
     search_opts = {'all_tenants': args.all_tenants}
 
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     groups = cs.groups.list(search_opts=search_opts)
 
     columns = ['ID', 'Status', 'Name']
     utils.print_list(groups, columns)
 
+    with cs.groups.completion_cache(
+            'uuid',
+            cinderclient.v3.groups.Group,
+            mode='w'):
+        for group in groups:
+            cs.groups.write_to_completion_cache('uuid', group.id)
+    with cs.groups.completion_cache('name',
+                                    cinderclient.v3.groups.Group,
+                                    mode='w'):
+        for group in groups:
+            if group.name is None:
+                continue
+            cs.groups.write_to_completion_cache('name', group.name)
+    AppendFilters.filters = []
+
 
 @api_versions.wraps('3.13')
 @utils.arg('--list-volume',
@@ -1415,6 +1438,17 @@ def do_group_create(cs, args):
     info.pop('links', None)
     utils.print_dict(info)
 
+    with cs.groups.completion_cache('uuid',
+                                    cinderclient.v3.groups.Group,
+                                    mode='a'):
+        cs.groups.write_to_completion_cache('uuid', group.id)
+
+    if group.name is not None:
+        with cs.groups.completion_cache('name',
+                                        cinderclient.v3.groups.Group,
+                                        mode='a'):
+            cs.groups.write_to_completion_cache('name', group.name)
+
 
 @api_versions.wraps('3.14')
 @utils.arg('--group-snapshot',
@@ -1616,6 +1650,7 @@ def do_group_list_replication_targets(cs, args):
            help="Filters results by a group ID. Default=None. "
                 "%s" % FILTER_DEPRECATED)
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -1633,13 +1668,14 @@ def do_group_snapshot_list(cs, args):
         'group_id': args.group_id,
     }
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     group_snapshots = cs.group_snapshots.list(search_opts=search_opts)
 
     columns = ['ID', 'Status', 'Name']
     utils.print_list(group_snapshots, columns)
+    AppendFilters.filters = []
 
 
 @api_versions.wraps('3.14')
@@ -1866,6 +1902,7 @@ def do_revert_to_snapshot(cs, args):
            help="Filters results by the message level. Default=None. "
                 "%s" % FILTER_DEPRECATED)
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -1882,8 +1919,8 @@ def do_message_list(cs, args):
         'request_id': args.request_id,
     }
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
     if args.resource_type:
         search_opts['resource_type'] = args.resource_type.upper()
     if args.level:
@@ -1905,6 +1942,7 @@ def do_message_list(cs, args):
     else:
         sortby_index = 0
     utils.print_list(messages, columns, sortby_index=sortby_index)
+    AppendFilters.filters = []
 
 
 @api_versions.wraps("3.3")
@@ -2004,6 +2042,7 @@ def do_message_delete(cs, args):
                 "volume api version >=3.22. Default=None. "
                 "%s" % FILTER_DEPRECATED)
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -2049,8 +2088,8 @@ def do_snapshot_list(cs, args):
     }
 
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     total_count = 0
     if show_count:
@@ -2086,6 +2125,7 @@ def do_snapshot_list(cs, args):
             mode='w'):
         for snapshot in snapshots:
             cs.volume_snapshots.write_to_completion_cache('uuid', snapshot.id)
+    AppendFilters.filters = []
 
 
 @api_versions.wraps('3.27')
@@ -2131,6 +2171,7 @@ def do_snapshot_list(cs, args):
            metavar='<tenant>',
            help='Display information from single tenant (Admin only).')
 @utils.arg('--filters',
+           action=AppendFilters,
            type=six.text_type,
            nargs='*',
            start_version='3.33',
@@ -2148,8 +2189,8 @@ def do_attachment_list(cs, args):
         'volume_id': args.volume_id,
     }
     # Update search option with `filters`
-    if hasattr(args, 'filters') and args.filters is not None:
-        search_opts.update(shell_utils.extract_filters(args.filters))
+    if AppendFilters.filters:
+        search_opts.update(shell_utils.extract_filters(AppendFilters.filters))
 
     attachments = cs.attachments.list(search_opts=search_opts,
                                       marker=args.marker,
@@ -2163,6 +2204,7 @@ def do_attachment_list(cs, args):
     else:
         sortby_index = 0
     utils.print_list(attachments, columns, sortby_index=sortby_index)
+    AppendFilters.filters = []
 
 
 @api_versions.wraps('3.27')
@@ -2506,3 +2548,41 @@ def do_transfer_create(cs, args):
 
     info.pop('links', None)
     utils.print_dict(info)
+
+
+@utils.arg('--all-tenants',
+           dest='all_tenants',
+           metavar='<0|1>',
+           nargs='?',
+           type=int,
+           const=1,
+           default=0,
+           help='Shows details for all tenants. Admin only.')
+@utils.arg('--all_tenants',
+           nargs='?',
+           type=int,
+           const=1,
+           help=argparse.SUPPRESS)
+@utils.arg('--sort',
+           metavar='<key>[:<direction>]',
+           default=None,
+           help='Sort keys and directions in the form of <key>[:<asc|desc>].',
+           start_version='3.59')
+def do_transfer_list(cs, args):
+    """Lists all transfers."""
+    all_tenants = int(os.environ.get("ALL_TENANTS", args.all_tenants))
+    search_opts = {
+        'all_tenants': all_tenants,
+    }
+
+    sort = getattr(args, 'sort', None)
+    if sort:
+        sort_args = sort.split(':')
+        if len(sort_args) > 2:
+            raise exceptions.CommandError(
+                'Invalid sort parameter provided. Argument must be in the '
+                'form "key[:<asc|desc>]".')
+
+    transfers = cs.transfers.list(search_opts=search_opts, sort=sort)
+    columns = ['ID', 'Volume ID', 'Name']
+    utils.print_list(transfers, columns)
